# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from . import minblep, pitch
from .budgie import readbytecode
from .clock import ClockInfo
from .config import ConfigName
from .dosound import Bytecode
from .iface import Config
from .lurlene import loadcontext, LurleneBridge, YM2149Chip
from .out import StereoInfo, YMStream
from .timerimpl import ChipTimer, SimpleChipTimer, SyncTimer
from .util import initlogging, MainThread
from .vis import NullRoll, RollImpl
from .ym2149 import LogicalRegisters, PhysicalRegisters, YM2149
from .ymformat import YMOpen
from .ymplayer import Player, LogicalBundle, PhysicalBundle
from diapyr import DI, types
from diapyr.start import Started
from lurlene.context import Context
import logging, lurlene.osc, sys

log = logging.getLogger(__name__)

def boot(configname):
    di = DI()
    di.add(configname.loadconfig)
    di.add(di)
    config = di(Config)
    config_repr = list(config.repr)
    if config_repr:
        for key in config_repr:
            print(repr((-getattr(config, key)).context().unravel()))
        sys.exit()
    di.add(ClockInfo)
    di.add(StereoInfo)
    di.add(YM2149)
    di.add(PhysicalRegisters)
    di.add(LogicalRegisters)
    di.add(minblep.loadorcreate)
    di.add(YMStream)
    pitch.configure(di)
    di.add(Context)
    di.add(RollImpl if config.pianorollenabled else NullRoll)
    di.add(MainThread)
    di.add(YM2149Chip)
    if config.SID.enabled:
        from . import resid
        resid.configure(di)
    return config, di

def main_bpmtool():
    'Show a table of speed (updates per tracker line) to BPM.'
    config, _ = boot(ConfigName())
    ups = config.updaterate
    lpb = config.linesperbeat
    for upl in range(1, 21):
        lpm = 60 * ups / upl
        bpm = lpm / lpb
        print(f"{upl:2} {bpm:7.3f}")

@types(Config, this = Bytecode)
def srcbytecodefactory(config):
    with open(config.inpath) as f:
        return Bytecode(readbytecode(f, config.srclabel), config.dosoundextraseconds)

def main_dosound2jack():
    'Play a Dosound script via JACK.'
    from . import jackclient
    initlogging()
    config, di = boot(ConfigName('inpath', 'srclabel'))
    with di:
        di.add(srcbytecodefactory)
        jackclient.configure(di)
        di.add(SyncTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_dosound2txt(): # TODO: Additional seconds not needed.
    'Render a Dosound script to logging.'
    from . import txt
    initlogging()
    config, di = boot(ConfigName('inpath', 'srclabel', name = 'txt'))
    with di:
        di.add(srcbytecodefactory)
        txt.configure(di)
        di.add(SimpleChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_dosound2wav():
    'Render a Dosound script to WAV.'
    from . import out
    initlogging()
    config, di = boot(ConfigName('inpath', 'srclabel', 'outpath'))
    with di:
        di.add(srcbytecodefactory)
        out.configure(di)
        di.add(ChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

@types(Config, this = Bytecode)
def dsdbytecodefactory(config):
    with open(config.inpath, 'rb') as f:
        log.debug("Total ticks: %s", (ord(f.read(1)) << 8) | ord(f.read(1)))
        return Bytecode(f.read(), config.dosoundextraseconds)

def main_dsd2wav():
    'Render Dosound bytecode to WAV.'
    from . import out
    initlogging()
    config, di = boot(ConfigName('inpath', 'outpath', name = 'dsd'))
    with di:
        di.add(dsdbytecodefactory)
        out.configure(di)
        di.add(ChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_lc2jack():
    'Play a Lurlene song via JACK.'
    from . import jackclient
    initlogging()
    config, di = boot(ConfigName('inpath', '--section'))
    with di:
        di.add(loadcontext)
        di.add(LurleneBridge)
        lurlene.osc.configure(di)
        di.add(SyncTimer)
        di.add(LogicalBundle)
        jackclient.configure(di)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_lc2portaudio():
    'Play a Lurlene song via PortAudio.'
    from . import portaudioclient
    initlogging()
    config, di = boot(ConfigName('inpath', '--section'))
    with di:
        di.add(loadcontext)
        di.add(LurleneBridge)
        lurlene.osc.configure(di)
        di.add(SyncTimer)
        di.add(LogicalBundle)
        portaudioclient.configure(di)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_lc2txt(args = sys.argv[1:]):
    'Render a Lurlene song to logging.'
    from . import txt
    initlogging()
    config, di = boot(ConfigName('inpath', '--section', name = 'txt', args = args))
    with di:
        di.add(loadcontext)
        di.add(LurleneBridge)
        txt.configure(di)
        di.add(SimpleChipTimer)
        di.add(LogicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_lc2wav(args = sys.argv[1:]):
    'Render a Lurlene song to WAV.'
    from . import out
    initlogging()
    config, di = boot(ConfigName('inpath', '--section', 'outpath', args = args))
    with di:
        di.add(loadcontext)
        di.add(LurleneBridge)
        out.configure(di)
        di.add(ChipTimer)
        di.add(LogicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_ym2jack():
    'Play a YM file via JACK.'
    from . import jackclient
    initlogging()
    config, di = boot(ConfigName('inpath'))
    with di:
        di.add(YMOpen)
        jackclient.configure(di)
        di.add(SyncTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_ym2portaudio():
    'Play a YM file via PortAudio.'
    from . import portaudioclient
    initlogging()
    config, di = boot(ConfigName('inpath'))
    with di:
        di.add(YMOpen)
        portaudioclient.configure(di)
        di.add(SyncTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_ym2txt():
    'Render a YM file to logging.'
    from . import txt
    initlogging()
    config, di = boot(ConfigName('inpath', name = 'txt'))
    with di:
        di.add(YMOpen)
        txt.configure(di)
        di.add(ChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()

def main_ym2wav():
    'Render a YM file to WAV.'
    from . import out
    initlogging()
    config, di = boot(ConfigName('inpath', 'outpath'))
    with di:
        di.add(YMOpen)
        out.configure(di)
        di.add(ChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Config
from argparse import ArgumentParser
from aridity import NoSuchPathException
from aridity.config import ConfigCtrl
from aridity.model import wrap
from diapyr import DI, types, UnsatisfiableRequestException
from importlib import import_module
from pathlib import Path
import logging, sys

log = logging.getLogger(__name__)

class ConfigName:

    namespace = 'pym2149'

    def __init__(self, *params, args = sys.argv[1:], name = 'defaultconf'):
        parser = ArgumentParser()
        parser.add_argument('--repr', action = 'append', default = [])
        parser.add_argument('--config', action = 'append', default = [])
        parser.add_argument('--ignore-settings', action = 'store_true')
        for param in params:
            parser.add_argument(param)
        self.additems = parser.parse_args(args)
        self.path = Path(__file__).resolve().parent / f"{name}.arid"

    @types(DI, this = Config)
    def loadconfig(self, di):
        config = ConfigCtrl()
        config.put('enter', function = enter)
        config.put('py', function = lambda *args: py(getattr(config.node, self.namespace), *args))
        config.put('resolve', function = lambda *args: AsContext.resolve(di, *args))
        config.printf("cwd = %s", self.path.parent)
        config.printf("%s . %s", self.namespace, self.path.name)
        if not self.additems.ignore_settings:
            try:
                config.loadsettings()
            except FileNotFoundError as e:
                log.warning("Settings not found: %s", e)
        for name, value in self.additems.__dict__.items():
            if 'config' == name:
                with config.repl() as repl:
                    for text in value:
                        repl.printf("%s", self.namespace)
                        for line in text.splitlines():
                            repl(f"\t{line}")
            else:
                setattr(getattr(config.node, self.namespace), name, value)
        return getattr(config.node, self.namespace)

class AsContext:

    @classmethod
    def resolve(cls, di, context, resolvable):
        try:
            return cls(context, di(_getglobal(context, resolvable).scalar))
        except UnsatisfiableRequestException:
            raise NoSuchPathException

    def __init__(self, parent, obj):
        self.parent = parent
        self.obj = obj

    def resolved(self, name):
        try:
            return wrap(getattr(self.obj, name))
        except AttributeError:
            return self.parent.resolved(name)

def _getglobal(context, resolvable):
    spec = resolvable.resolve(context).cat()
    lastdot = spec.rindex('.')
    return wrap(getattr(import_module(spec[:lastdot], __package__), spec[lastdot + 1:]))

def enter(context, contextresolvable, resolvable):
    return resolvable.resolve(contextresolvable.resolve(context))

def py(config, context, *clauses):
    return wrap(eval(' '.join(c.cat() for c in clauses), dict(config = config)))



# Copyright 2014, 2018, 2019, 2020 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .const import clientname
from .iface import AmpScale, Config, Platform, Stream
from .nod import Node
from .out import FloatStream, StereoInfo
from diapyr import types
import outjack.jackclient as jc, logging

log = logging.getLogger(__name__)

class JackClient(jc.JackClient, Platform):

    @types(Config, StereoInfo)
    def __init__(self, config, stereoinfo):
        portcount = stereoinfo.getoutchans.size + config.SID.enabled # FIXME: This is a hack.
        super().__init__(clientname, portcount, config.jackringsize, config.jackcoupling)

    def start(self):
        super().start()
        log.debug(
            "JACK block size: %s or %.3f seconds",
            self.buffersize,
            self.buffersize / self.outputrate)

class JackStream(Stream, Node, metaclass = AmpScale):

    log2maxpeaktopeak = 1

    @types(Config, [FloatStream], JackClient)
    def __init__(self, config, streams, client):
        super().__init__()
        self.systemchannelcount = config.systemchannelcount
        for stream in streams:
            for chanindex in range(stream.chancount):
                client.port_register_output(f"{stream.streamname}_{1 + chanindex}")
        self.streams = streams
        self.client = client

    def start(self):
        self.client.activate()
        for stream in self.streams:
            # Connect all system channels, cycling over our streams if necessary:
            for syschanindex in range(self.systemchannelcount):
                chanindex = syschanindex % stream.chancount
                self.client.connect(f"{clientname}:{stream.streamname}_{1 + chanindex}", f"system:playback_{1 + syschanindex}")
        self.filler = BufferFiller(sum(s.chancount for s in self.streams), self.client.buffersize, self.client.current_output_buffer, self.client.send_and_get_output_buffer)

    def callimpl(self):
        self.filler([self.chain(wav) for stream in self.streams for wav in stream])

    def flush(self):
        pass # Nothing to be done.

    def stop(self):
        self.client.deactivate()

class BufferFiller:

    def __init__(self, portcount, buffersize, init, flip):
        self.portcount = portcount
        self.buffersize = buffersize
        self._newbuf(init)
        self.flip = flip

    def __call__(self, outbufs):
        n = len(outbufs[0])
        i = 0
        while i < n:
            m = min(n - i, self.buffersize - self.cursor)
            for portindex in range(self.portcount):
                self.outbuf[portindex, self.cursor:self.cursor + m] = outbufs[portindex].buf[i:i + m]
            self.cursor += m
            i += m
            if self.cursor == self.buffersize:
                self._newbuf(self.flip)

    def _newbuf(self, factory):
        outbuf = factory().view()
        outbuf.shape = self.portcount, self.buffersize
        self.outbuf = outbuf
        self.cursor = 0

def configure(di):
    di.add(JackClient)
    di.add(JackStream)
